/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#include "buffer.h"
#include "utility.h"


#include <assert.h>

#include <iostream>
#include <iomanip>
#include <algorithm>
#include <error.h>

namespace bit
  {

  const uint8_t masks[8] =
    {
      0x00, 0x01, 0x03, 0x07, 0x0F, 0x1F, 0x3F, 0x7F
    };

  Buffer::Buffer(size_t size, Growth growth, size_t sizemax):
      m_pdata(NULL),
      m_size(0),
      m_is_owner(true),
      m_growth(growth),
      m_sizemax(sizemax),
      m_signal_data_changed_need_reemit(false),
      m_signal_data_changed_emitting(false)
  {
    set_size(size);
  }

  Buffer::Buffer( void * external_data, size_t data_octets, DataMode datamode, Growth growth, size_t sizemax ):
      m_pdata(NULL),
      m_size(0),
      m_is_owner(true),
      m_growth(growth),
      m_sizemax(sizemax)
  {
    set_data(external_data, data_octets, datamode);
  }

  Buffer::~Buffer()
  {
    if (m_is_owner)
      free_buffer();
  }

  const uint8_t * const Buffer::data() const
    {
      return m_pdata;
    }

  size_t Buffer::size() const
    {
      return m_size;
    }

  bool Buffer::is_dynamic() const
    {
      return m_growth == DYNAMIC;
    }

  bool Buffer::is_fixed() const
    {
      return m_growth == FIXED;
    }

  Growth Buffer::growth() const
    {
      return m_growth;
    }

  void Buffer::set_growth(Growth growth)
  {
    if ( m_is_owner )
      m_growth = growth;
  }

  /** @todo optimize for the START case */
  bool Buffer::unpack(void* mem, size_t octets, Location mem_loc, size_t offset, size_t extract)
  {
    size_t ot, bsd, befpo, bsu, fwo, lwo, po;
    uint8_t mask, temp;
    uint8_t *ad, *p;
    size_t needed_size;

    // make basic assumptions: do we have a data buffer? are we extracting into a null address?
    if (   m_pdata == NULL || mem == NULL )
      return false;

    // are we being asked to extract more octets than the target can accept
    if ( octets < octet_ceiling(extract) )
      // if so, reduce the request size
      extract = octets * 8;

    // check to see if the request is beyond the current size
    if ( m_size < octet_ceiling(offset+extract) )
      {
        if (! this->is_dynamic() )
          return false; // there's nothing we can do if we aren't dynamic

        // try to adjust the buffer size to accomodate the request
        needed_size = octet_ceiling(offset+extract);
        if (set_size(needed_size) < needed_size)
          return false; // the buffer resize request failed
      }

    // clear target memory
    memset(mem, 0x00, octets);

    if (offset%8 && (offset+extract)%8 && offset/8==(offset+extract)/8)
      {
        uint8_t mask, value;
        mask = masks[last_octet_lower_bits(offset)] ^ masks[last_octet_lower_bits(offset+extract)];
        value = *(m_pdata+offset/8) & mask;
        value >>= last_octet_lower_bits(offset+extract);
        *((uint8_t*)mem+octets-1) = value;
      }
    else
      {
        // first whole octet to transfer is integer divisor plus 1 for any remainder
        fwo = octet_ceiling(offset);
        // calculate last whole octet
        lwo = octet_ceiling(offset+extract);
        // calculate octets to transfer
        ot = lwo - fwo;
        // calculate actual destination within memory area
        ad = (uint8_t*)mem + octets - ot;
        // and perform memory copy (if we have anything to copy)
        // if we only have a first partial, then there won't be a copy here
        // and it will occur later when we patch up the front
        if (ot)
          memcpy(ad, m_pdata+fwo, ot);

        // downward shift on chunk
        bsd = last_octet_lower_bits(offset+extract);
        if (bsd)
          {
            mask = masks[bsd];
            // calculate bits to shift upward in each chunk
            bsu = 8 - bsd;
            for (p = ad+ot-1; p > ad; p--)
              {
                *p >>= bsd;
                temp = *(p-1) & mask;
                temp <<= bsu;
                *p |= temp;
              }
            // don't forget the uppermost byte which only needs to be shifted
            *p >>= bsd;
          }

        befpo = last_octet_lower_bits(offset);
        if (befpo)
          {
            po = offset / 8;
            if (bsd)
              {
                // error somewhere in here
                temp = *( m_pdata + po );
                temp &= masks[befpo];
                temp <<= bsd;
                *(ad+fwo-1) |= temp;
              }
            if (befpo > bsd)
              {
                temp = *(m_pdata+po);
                temp &= masks[befpo];
                temp >>= bsd;
                *(ad-1) = temp;
              }
          }
      }
    if (mem_loc == START)
        left_shift(mem, octets, octets*8-extract);
    return true;
  }

  /** @todo optimize for the START case */
  bool Buffer::pack(const void* mem, size_t octets, Location mem_loc, size_t offset, size_t destsize, size_t n)
  {
    uint8_t *pbuf, *pmem;
    uint8_t *tmem = static_cast<uint8_t*>( const_cast<void*>(mem) );
    uint8_t temp, temp1, temp2, mask;
    size_t copy;
    size_t start, end, lb, sb, eb, rb, bits;
    size_t needed_size;

    // are we extracting from a null address?
    if ( mem == NULL )
      return false;

    // are we being asked to pack in more bits than will fit?
    if (n > destsize)
      n = destsize;

    // check to see if the request is beyond the current size
    if ( m_size < octet_ceiling(offset+destsize))
      {
        if ( ! this->is_dynamic() )
          return false; // there's nothing we can do if we aren't dynamic

        // try to adjust the buffer size to accomodate the request
        needed_size = octet_ceiling(offset+destsize);
        if (set_size(needed_size) < needed_size)
          return false; // the buffer resize request failed
      }

    start = offset + destsize - n;
    end = offset + destsize;

    // HACK
    if (mem_loc == START) {
      tmem = static_cast<uint8_t*>( malloc(octets) );
      memcpy(tmem, mem, octets);
      right_shift(tmem, octets, octets*8-n);
    }
    // END HACK
    
    // does buffer target end on octet boundary?
    if ( end%8 == 0)
      { // ends on octet boundary
        // copy as much as possible
        copy = n/8; // calculate whole octets to copy
        pbuf = m_pdata + end/8 - copy; // calculate buffer location by moving forward to end and backing up # octets to copy
        pmem = tmem + octets - copy; // calculate copy location in memory similar to pbuf
        memcpy(pbuf, pmem, copy);

        // copy any remaining bits
        if (n%8)
          { // check to see if there are remaining bits
            pbuf--;
            pmem--;
            temp1 = *(pmem);
            temp1 &= masks[n%8];
            temp2 = *(pbuf);
            temp2 &= ~masks[n%8];
            *(pbuf) = temp1 | temp2;
          }

      }
    // check to see if we have a partial octet, which we have if start and end are in the same octet
    else if (start/8 == end/8)
      {
        bits = (end-start)%8; // calculate how many bits we will extract
        mask = masks[bits]; // get the bit mask to use
        lb = last_octet_lower_bits(end); // calculate the shift offset
        temp = *(tmem+octets-1); // get last octet of memory
        temp <<= lb; // shift up into position
        mask <<= lb; // shift mask up into position as well
        temp &= mask; // mask out all bits not copied
        *(m_pdata+start/8) &= ~mask; // mask out the bits to be copied in destination
        *(m_pdata+start/8) |= temp; // and finally bring in our current bits
      }
    else
      { // we don't end on an octet boundary and cross at least one octet boundary
        // jump to the last octet
        pmem = tmem + octets - 1;
        sb = last_octet_upper_bits(end); // calculate the number of shift bits to keep
        lb = 8 - sb; // calculate the number of lower bits to drop
        pbuf = m_pdata + end/8;
        *pbuf &= masks[lb]; // clear upper bits
        *pbuf |= *pmem << lb; // bring in upper bits
        pbuf--; // start with pbuf at first whole octet, but memory still at first octet
        while (pbuf > m_pdata + start/8)
          {
            *pbuf = *pmem >> sb; // bring in upper bits of memory into lower bits of buffer
            pmem--; // move back one octet on memory
            *pbuf |= *pmem << lb; // and now lower bits of memory into upper bits of buffer
            pbuf--; // and back up one octet in the buffer
          }
        // we just finished the last whole octet, now we need to do the leading octet
        if (start%8 == 0)
          {
            *pbuf = *pmem >> sb; // bring in upper bits of memory into lower bits of buffer
            pmem--; // move back one octet on memory
            *pbuf |= *pmem << lb; // and now lower bits of memory into upper bits of buffer
          }
        else
          { // check to see if there is any work to be done
            rb = last_octet_lower_bits(start); // remaining bits for leading octet
            // use as many bits as we can from the current memory pointer
            mask = masks[((rb>lb)?lb:rb)]; // get mask which is lesser of lb or sb
            temp = *pbuf;
            *pbuf &= ~mask; // clear bits for lesser of rb or lb
            temp = *pbuf;
            *pbuf |= (*pmem >> sb) & mask; // get bits shifted down and potentially clear for rb mask
            temp = *pbuf;
            if (rb > lb)
              { // rb was the greater so we need to get rb-lb more bits from memory
                eb = rb - lb;
                pmem--; // move back one more octet
                mask = mask ^ masks[rb]; // get mask which represents only the extra bits
                temp = *pbuf;
                *pbuf &= ~mask; // clear the extra bits
                temp = *pbuf;
                temp = *pmem;
                *pbuf |= (*pmem << lb) & mask; // shift memory up and mask only needed bits
                temp = *pbuf;
              }
          }
      }

    // zero out any leading data in buffer
    if (destsize > n)
      clear_bits(offset, destsize-n, true);

    // HACK
    if (mem_loc == START)
      free(tmem);
    // END HACK

    on_data_changed();
    return true;
  }

  void Buffer::set_data(void* data, size_t external_size, DataMode datamode)
  {
    switch (datamode)
      {
      case COPY:
        if ( ! m_is_owner )
          {
            m_pdata = NULL;
            m_size = 0;
            m_is_owner = true;
          }

        set_size(external_size);

        memcpy(m_pdata, data, std::min(external_size,m_size) );
        if (external_size < m_size)
          memset(m_pdata+external_size, 0x00, m_size-external_size);
        on_data_changed();
        break;
      case TAKE_OWNERSHIP:
        if (m_is_owner)
          free_buffer();
        m_pdata = static_cast<uint8_t*>(data);
        m_size = external_size;
        if (m_sizemax < m_size)
          m_sizemax = m_size;
        m_is_owner = true;
        break;
      case EXTERNAL_OWNER:
        if (m_is_owner)
          free_buffer();
        m_pdata = static_cast<uint8_t*>(data);
        m_size = external_size;
        m_sizemax = m_size;
        m_is_owner = false;
        m_growth = FIXED;
        break;
      default:
        throw std::logic_error("bit:Buffer Data mode not implemented");
      }
  }

  void Buffer::clear()
  {
    memset(m_pdata, 0x00, m_size);
    on_data_changed();
  }

  bool Buffer::clear_bits(size_t offset, size_t bits, bool suppress)
  {
    if (m_pdata == NULL || octet_ceiling(offset+bits) > m_size)
      return false;
    if (offset%8 && (offset+bits)%8 && offset/8==(offset+bits)/8)
      {
        uint8_t mask;
        mask = masks[last_octet_lower_bits(offset)] ^ masks[last_octet_lower_bits(offset+bits)];
        *(m_pdata+offset/8) &= ~mask;
      }
    else
      {
        // clear the leading portion (if any)
        if ( last_octet_lower_bits(offset) )
          *( static_cast<uint8_t*>(starting_octet( m_pdata, offset)) ) &= ~(masks[last_octet_lower_bits(offset)]);
        // clear any whole octets (if any)
        if ( ending_octet( m_pdata, offset) != starting_octet( m_pdata, offset+bits) )
          memset(ending_octet( m_pdata, offset), 0x00, static_cast<uint8_t*>(starting_octet( m_pdata, offset+bits))-static_cast<uint8_t*>(ending_octet( m_pdata, offset)));
        // clear the trailing portion (if any)
        if ( last_octet_upper_bits(offset+bits) )
          *( static_cast<uint8_t*>(starting_octet( m_pdata, offset+bits)) ) &= masks[last_octet_lower_bits(offset+bits)];
      }

    if (!suppress)
      on_data_changed();

    return true;
  }

  Buffer& Buffer::operator=( const Buffer & other )
  {
    // the dynamic and owner conditions will always be the same as the other buffer
    m_growth = other.m_growth;
    m_sizemax = other.m_sizemax;

    // duplicate buffer
    set_size(other.m_size);
    memcpy(m_pdata, other.m_pdata, m_size);
    on_data_changed();

    return *this;
  }

  size_t Buffer::set_size( size_t size_request )
  {
    uint8_t* new_buffer;
    bool location_changed = false;

    if (! this->is_dynamic() )
      return m_size;

    if (! m_is_owner )
      return m_size;

    // a request of size 0 is a request to free the buffer
    if (size_request == 0)
      {
        free_buffer();
        on_data_location_changed();
        on_size_changed();
        return m_size;
      }

    // make sure we don't make it larger than the max size
    // if a maximum size is set
    if (m_sizemax > 0)
      size_request = std::min(size_request, m_sizemax);

    new_buffer = (uint8_t*) realloc(m_pdata, size_request);
    // check to make sure the allocation went through
    if (new_buffer != NULL)
      {
        // check to see if the location changed, but if so, wait until m_pdata and
        // m_size are changed before sending notifications
        if (m_pdata != new_buffer)
          location_changed = true;
        m_pdata = new_buffer;
        // clear any new memory if there was any
        if (size_request > m_size)
          memset(m_pdata+m_size, 0x00, size_request-m_size);
        m_size = size_request;
        on_data_location_changed();
        on_size_changed();
      }

    return m_size;
  }

  void Buffer::free_buffer( )
  {
    assert( m_is_owner );
    if (m_pdata != NULL)
      free(m_pdata);
    m_pdata = NULL;
    m_size = 0;
  }

  size_t Buffer::sizemax( )
  {
    return m_sizemax;
  }

  void Buffer::set_sizemax(size_t sizemax)
  {
    m_sizemax = sizemax;
    if (m_sizemax < m_size)
      {
        set_size(m_sizemax);
      }
  }

  void Buffer::on_data_location_changed( )
  {
    m_signal_data_location_changed();
  }

  void Buffer::on_size_changed( )
  {
    m_signal_size_changed();
  }

  void Buffer::on_data_changed( )
  {
    if (m_signal_data_changed_emitting)
      {
        m_signal_data_changed_need_reemit = true;
        return;
      }
    m_signal_data_changed_emitting = true;
    do
      {
        m_signal_data_changed();
      }
    while (m_signal_data_changed_need_reemit);
    m_signal_data_changed_need_reemit = false;
    m_signal_data_changed_emitting = false;
    return;
  }

}
